from pathlib import Path
from typing import Dict, Optional, Set, Tuple

from phiterm.conf.phi_conf import PhiWsData
from phiterm.utils.cli_console import (
    print_info,
    print_info,
    print_error,
    print_subheading,
)
from phiterm.utils.log import logger
from phiterm.utils.pyproject import read_pyproject_phidata
from phiterm.workspace.ws_enums import WorkspaceSetupActions


def get_ws_config_dir_path(ws_root: Path) -> Path:
    logger.debug(f"Searching for a workspace directory in {ws_root}")
    # Phidata workspace dir can be found at:
    # 1. sub dir: workspace
    # 2. In a folder defined by the pyproject.toml file

    # Case 1: Look for a folder with name: workspace
    ws_workspace_dir = ws_root.joinpath("workspace")
    logger.debug(f"Searching {ws_workspace_dir}")
    if ws_workspace_dir.exists() and ws_workspace_dir.is_dir():
        return ws_workspace_dir
        # ws_config_file = ws_workspace_dir.joinpath("config.py")
        # if (
        #     ws_config_file is not None
        #     and ws_config_file.exists()
        #     and ws_config_file.is_file()
        # ):
        #     return ws_workspace_dir

    # Case 2: Look for a folder defined by the pyproject.toml file
    ws_pyproject_toml = ws_root.joinpath("pyproject.toml")
    if ws_pyproject_toml.exists() and ws_pyproject_toml.is_file():
        phidata_conf = read_pyproject_phidata(ws_pyproject_toml)
        if phidata_conf is not None:
            phidat_conf_workspace_dir_str = phidata_conf.get("workspace", None)
            phidat_conf_workspace_dir = ws_root.joinpath(phidat_conf_workspace_dir_str)
            logger.debug(f"searching {phidat_conf_workspace_dir}")
            if (
                phidat_conf_workspace_dir.exists()
                and phidat_conf_workspace_dir.is_dir()
            ):
                return phidat_conf_workspace_dir
                # ws_config_file = phidat_conf_workspace_dir.joinpath("config.py")
                # if (
                #     ws_config_file is not None
                #     and ws_config_file.exists()
                #     and ws_config_file.is_file()
                # ):
                #     return phidat_conf_workspace_dir

    print_error(f"Could not find a workspace dir for {ws_root}")
    exit(0)


def get_ws_config_file_path(ws_root: Path) -> Optional[Path]:
    logger.debug(f"Looking for a workspace config for {ws_root}")
    # Phidata config.py file can be found at:
    # 1. In a folder named workspace
    # 2. In a folder defined by the pyproject.toml file

    # Case 1: Look for a folder with name: workspace
    ws_workspace_dir = ws_root.joinpath("workspace")
    logger.debug(f"Searching {ws_workspace_dir}")
    if ws_workspace_dir.exists() and ws_workspace_dir.is_dir():
        ws_config_file = ws_workspace_dir.joinpath("config.py")
        if (
            ws_config_file is not None
            and ws_config_file.exists()
            and ws_config_file.is_file()
        ):
            return ws_config_file

    # Case 2: Look for a folder defined by the pyproject.toml file
    ws_pyproject_toml = ws_root.joinpath("pyproject.toml")
    if ws_pyproject_toml.exists() and ws_pyproject_toml.is_file():
        phidata_conf = read_pyproject_phidata(ws_pyproject_toml)
        if phidata_conf is not None:
            phidat_conf_workspace_dir_str = phidata_conf.get("workspace", None)
            phidat_conf_workspace_dir = ws_root.joinpath(phidat_conf_workspace_dir_str)
            logger.debug(f"Searching {phidat_conf_workspace_dir}")
            if (
                phidat_conf_workspace_dir.exists()
                and phidat_conf_workspace_dir.is_dir()
            ):
                ws_config_file = phidat_conf_workspace_dir.joinpath("config.py")
                if (
                    ws_config_file is not None
                    and ws_config_file.exists()
                    and ws_config_file.is_file()
                ):
                    return ws_config_file

    return None


def get_ws_k8s_resources_dir_path(ws_root: Path) -> Path:
    return ws_root.joinpath("workspace").joinpath("prd")


def is_valid_ws_config_file_path(ws_config_file_path: Optional[Path]) -> bool:
    # TODO: add more checks for validating the ws_config_file_path
    if (
        ws_config_file_path is not None
        and ws_config_file_path.exists()
        and ws_config_file_path.is_file()
    ):
        return True
    return False


def get_ws_setup_status(ws_data: PhiWsData) -> Tuple[bool, str]:
    """This function validates that a workspace is properly set up and returns an
    error message if ws not setup. The error message is displayed to the user,
    so we need to make sure it doesn't contain any sensitive information.
    """

    # Version Control Provider check
    # user: UserSchema = config.user
    # if user.version_control_provider is None:
    #     return (
    #         False,
    #         "Version control provider not available. Please run `phi init -r` to initialize phi",
    #     )

    if ws_data is None:
        return (
            False,
            "WorkspaceSchema not yet registered with Phidata. Please run `phi ws init` to create a new workspace or `phi ws setup` for an existing workspace",
        )

    # Validate WorkspaceSchema Directory
    ws_root_path: Optional[Path] = ws_data.ws_root_path
    if ws_root_path is None or not ws_root_path.exists() or not ws_root_path.is_dir():
        return (
            False,
            "The WorkspaceSchema directory is not available or invalid.\n\tTo create a new workspace, run `phi ws init`.\n\tFor an existing workspace, run `phi ws setup` from the workspace directory",
        )

    # Validate WorkspaceSchema Config Exists (we will validate the contents later)
    ws_config_file_path: Optional[Path] = ws_data.ws_config_file_path
    if (
        ws_config_file_path is None
        or not ws_config_file_path.exists()
        or not ws_config_file_path.is_file()
    ):
        return (
            False,
            "WorkspaceSchema config is not available or invalid. Please run `phi ws init` to create a new workspace or `phi ws setup` for an existing workspace",
        )

    # Validate that the workspace is registered with phidata
    if ws_data is not None and ws_data.ws_schema is None:
        return (
            False,
            "WorkspaceSchema not registered with phidata. Please run `phi ws setup` from the workspace dir",
        )

    # Validate that the workspace has a git repo
    if (
        ws_data is not None
        and ws_data.ws_schema is not None
        # and is_empty(ws_data.ws_schema.git_url)
    ):
        return (
            False,
            "WorkspaceSchema does not have a remote origin setup. Please run `phi ws setup` from the workspace dir",
        )

    return (True, "WorkspaceSchema setup successful")


ws_setup_action_todo: Dict[WorkspaceSetupActions, str] = {
    WorkspaceSetupActions.WS_CONFIG_IS_AVL: "[ERROR] WorkspaceConfig is missing",
    WorkspaceSetupActions.WS_IS_AUTHENTICATED: "Workspace is not authenticated. run `phi auth` for more info",
    WorkspaceSetupActions.GCP_SVC_ACCOUNT_IS_AVL: "GCP Service Account is unavailable. run `phi gcp auth` for more info",
    WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL: "No git repo setup for workspace. run `phi ws git` for more info",
}


def print_howtofix_pending_actions(
    pending_actions: Optional[Set[WorkspaceSetupActions]],
) -> None:
    """This function prints how to fix pending setup actions for a workspace"""

    if pending_actions is None:
        return

    for action in pending_actions:
        print_info("\t" + ws_setup_action_todo.get(action, action.value))


def secho_ws_status(ws_data: PhiWsData) -> None:
    ws_root_path: Optional[Path] = ws_data.ws_root_path
    ws_config_file_path: Optional[Path] = ws_data.ws_config_file_path

    print_info("")
    print_subheading("WorkspaceSchema: {}".format(ws_data.ws_name))
    print_info("WorkspaceSchema Directory: {}".format(str(ws_root_path)))
    print_info("Workspace Config file: {}".format(str(ws_config_file_path)))

    ws_is_setup, ws_setup_msg = get_ws_setup_status(ws_data)
    if ws_is_setup:
        print_info("WorkspaceSchema Status: Active")
    else:
        print_info("WorkspaceSchema Status: {}".format(ws_setup_msg))

    # ws_pak8_conf: Optional[Pak8Conf] = ws_data.ws_pak8_conf
    # if ws_pak8_conf and ws_pak8_conf.cloud_provider == pak8_Pak8CloudProvider.GCP:
    #     secho_gcp_status(ws_data)



from typing import Any, Dict, List, Optional, Union

from httpx import Client, Response, NetworkError, codes

from phiterm.api.client import get_authenticated_client, base_headers
from phiterm.api.exceptions import CliAuthException
from phiterm.api.routes import ApiRoutes
from phiterm.api.handler import invalid_response
from phiterm.conf.constants import BACKEND_API_URL, PHI_AUTH_TOKEN_HEADER
from phiterm.schemas.user import UserSchema, EmailPasswordSignInSchema
from phiterm.utils.cli_console import log_network_error_msg
from phiterm.utils.log import logger


def user_ping() -> bool:
    logger.debug("--o-o-- Ping user api")

    with Client(base_url=BACKEND_API_URL, headers=base_headers, timeout=60) as api:
        try:
            r: Response = api.get(ApiRoutes.USER_PING)
            if invalid_response(r):
                return False
        except NetworkError as e:
            log_network_error_msg()
            return False

        if r.status_code == codes.OK:
            return True
    return False


def is_user_authenticated() -> bool:
    from phiterm.conf.phi_conf import PhiConf

    logger.debug("--o-o-- Authenticating user")

    authenticated_client: Optional[Client] = get_authenticated_client()
    if authenticated_client is None:
        logger.debug("Session not available")
        return False

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        return False
    user: Optional[UserSchema] = phi_conf.user
    if user is None:
        return False

    with authenticated_client as api:
        try:
            r: Response = api.post(
                ApiRoutes.USER_AUTHENTICATE, data=user.json(exclude_none=True)  # type: ignore
            )
            if invalid_response(r):
                return False
        except NetworkError:
            log_network_error_msg()
            return False

        response_data: Union[Dict[Any, Any], List[Any]] = r.json()
        if response_data is None or not isinstance(response_data, dict):
            logger.error("Could not parse response")
            return False
        if response_data.get("status", "fail") == "success":
            return True
    return False


def authenticate_and_get_user(
    tmp_auth_token: str,
) -> Optional[UserSchema]:
    from phiterm.conf.auth import save_auth_token

    logger.debug("--o-o-- Getting user")

    authenticated_headers = base_headers.copy()
    authenticated_headers[PHI_AUTH_TOKEN_HEADER] = tmp_auth_token
    with Client(
        base_url=BACKEND_API_URL,
        headers=authenticated_headers,
        timeout=None,
    ) as api:
        try:
            r: Response = api.post(
                ApiRoutes.USER_CLI_AUTH, data={"token": tmp_auth_token}
            )
            if invalid_response(r):
                return None
        except NetworkError:
            log_network_error_msg()
            return None

        phidata_auth_token = r.headers.get(PHI_AUTH_TOKEN_HEADER)
        if phidata_auth_token is None:
            logger.error("Could not authenticate user")
            return None

        user_data = r.json()
        if not isinstance(user_data, dict):
            raise CliAuthException("Could not parse response")

        current_user: UserSchema = UserSchema.from_dict(user_data)
        if current_user is not None:
            save_auth_token(phidata_auth_token)
            return current_user
    return None


def sign_in_user(
    sign_in_data: EmailPasswordSignInSchema,
) -> Optional[UserSchema]:
    from phiterm.conf.auth import save_auth_token

    logger.debug("--o-o-- Sign in user")

    with Client(
        base_url=BACKEND_API_URL,
        headers=base_headers,
        timeout=None,
    ) as api:
        try:
            r: Response = api.post(ApiRoutes.USER_SIGN_IN, json=sign_in_data.dict())
            if invalid_response(r):
                return None
        except NetworkError as e:
            log_network_error_msg()
            return None

        phidata_auth_token = r.headers.get(PHI_AUTH_TOKEN_HEADER)
        if phidata_auth_token is None:
            logger.error("Could not authenticate user")
            return None

        user_data = r.json()
        if not isinstance(user_data, dict):
            raise CliAuthException("Could not parse response")

        current_user: UserSchema = UserSchema.from_dict(user_data)
        if current_user is not None:
            save_auth_token(phidata_auth_token)
            return current_user
    return None



from phiterm.utils.enums import ExtendedEnum


class WorkspaceVisibilityEnum(ExtendedEnum):
    PRIVATE = "PRIVATE"
    PUBLIC = "PUBLIC"


class WorkspaceStarterTemplate(ExtendedEnum):
    ml_app = "ml-app"
    ai_app = "ai-app"
    duck_gpt = "duck-gpt"
    aws_dp = "data-platform"
    aws_spark_dp = "spark-data-platform"
    aws_snowflake_dp = "snowflake-data-platform"


class WorkspaceEnv(ExtendedEnum):
    dev = "dev"
    stg = "stg"
    prd = "prd"
    test = "test"
    all = "all"


class WorkspaceConfigType(ExtendedEnum):
    docker = "docker"
    k8s = "k8s"
    gcp = "gcp"
    aws = "aws"


class WorkspaceSetupActions(ExtendedEnum):
    WS_CONFIG_IS_AVL = "WS_CONFIG_IS_AVL"
    WS_IS_AUTHENTICATED = "WS_IS_AUTHENTICATED"
    GCP_SVC_ACCOUNT_IS_AVL = "GCP_SVC_ACCOUNT_IS_AVL"
    GIT_REMOTE_ORIGIN_IS_AVL = "GIT_REMOTE_ORIGIN_IS_AVL"



# from typing import Any, Dict, List, Optional
#
# import pydantic
#
# import pak8.enums as pak8_enums
# from pak8.conf import Pak8Conf
# from pak8.gcp.gcp_pak8 import GCPPak8
# from pak8.k8s.clients.pak8_k8s_client import Pak8K8sClient
# from phiterm.utils.common import is_empty
# from phiterm.utils.log import logger
#
#
# def get_k8s_resources_as_dicts(
#     pak8_conf: Pak8Conf,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> Optional[List[Dict[str, Any]]]:
#     """For a given Pak8Conf, gets the k8s manifests as Dictionaries and returns them
#     as a List of Dicts.
#
#     Args:
#     pak8_conf: Conf being deployed
#
#     Returns:
#     Optional[List[Dict[str, Any]]]: List of K8s resources which can be applied to a k8s cluster after converting to yaml
#     """
#
#     logger.debug("Creating Pak8: {}".format(pak8_conf.name))
#
#     # Step 1: Create a GCPPak8 from the pak8_conf and validate
#     ws_pak8: Optional[GCPPak8] = None
#     try:
#         if pak8_conf.cloud_provider == pak8_enums.Pak8CloudProvider.GCP:
#             ws_pak8 = GCPPak8(pak8_conf)
#             logger.debug("GCPPak8 Created")
#             if ws_pak8 is None or not isinstance(ws_pak8, GCPPak8):
#                 logger.error("Invalid GCPPak8: {}".format(ws_pak8))
#                 return None
#         else:
#             logger.error(
#                 "Cloud Provider {} not yet supported".format(pak8_conf.cloud_provider)
#             )
#             return None
#     except pydantic.ValidationError as e:
#         logger.debug("e: {}".format(e))
#         error_msg = "Config Validation Errors. Details:\n{}".format(e)
#         logger.error(error_msg)
#         return None
#
#     ws_k8s_resource = ws_pak8.get_k8s_resources_as_dicts(
#         type_filters=type_filters, name_filters=name_filters
#     )
#
#     return ws_k8s_resource
#
#
# def get_k8s_resource_groups_as_dicts(
#     pak8_conf: Pak8Conf,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> Optional[Dict[str, List[Dict[str, Any]]]]:
#     """For a given Pak8Conf, gets the k8s manifests as Dictionaries and returns them
#     as a List of Dicts.
#
#     Args:
#     pak8_conf: Conf being deployed
#
#     Returns:
#     Optional[List[Dict[str, Any]]]: List of K8s resources which can be applied to a k8s cluster after converting to yaml
#     """
#
#     logger.debug("Creating Pak8: {}".format(pak8_conf.name))
#
#     # Step 1: Create a GCPPak8 from the pak8_conf and validate
#     ws_pak8: Optional[GCPPak8] = None
#     try:
#         if pak8_conf.cloud_provider == pak8_enums.Pak8CloudProvider.GCP:
#             ws_pak8 = GCPPak8(pak8_conf)
#             logger.debug("GCPPak8 Created")
#             if ws_pak8 is None or not isinstance(ws_pak8, GCPPak8):
#                 logger.error("Invalid GCPPak8: {}".format(ws_pak8))
#                 return None
#         else:
#             logger.error(
#                 "Cloud Provider {} not yet supported".format(pak8_conf.cloud_provider)
#             )
#             return None
#     except pydantic.ValidationError as e:
#         logger.debug("e: {}".format(e))
#         error_msg = "Config Validation Errors. Details:\n{}".format(e)
#         logger.error(error_msg)
#         return None
#
#     ws_k8s_resource_groups: Optional[
#         Dict[str, List[Dict[str, Any]]]
#     ] = ws_pak8.get_k8s_resource_groups_as_dicts(
#         type_filters=type_filters, name_filters=name_filters
#     )
#
#     return ws_k8s_resource_groups
#
#
# def apply_k8s_resources_for_k8s_client(
#     pak8_conf: Pak8Conf,
#     pak8_k8s_client: Pak8K8sClient,
#     service_filters: Optional[List[str]] = None,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> bool:
#     """For a given Pak8Conf and Pak8K8sClient, applies the k8s manifests
#
#     Args:
#     pak8_conf: Conf being deployed
#     pak8_k8s_client:
#
#     Returns:
#     bool: True is the resources were applied successfully
#     """
#
#     if is_empty(service_filters):
#         pak8_k8s_client.create_resources(
#             type_filters=type_filters,
#             name_filters=name_filters,
#         )
#         return True
#
#     ######################################################
#     ## Applying Pak8 Services
#     ######################################################
#
#     # if pak8_conf.services is None:
#     #     return False
#     #
#     # # Create a list of services to apply
#     # _pak8_svc_dict = {s.name: s for s in pak8_conf.services}
#     # _pak8_services_to_apply: List[pak8.Pak8AppConf] = []
#     # for _svc_nm in service_filters:
#     #     if _svc_nm not in _pak8_svc_dict:
#     #         pprint_error(f"Service {_svc_nm} not found")
#     #         continue
#     #     _pak8_services_to_apply.append(_pak8_svc_dict[_svc_nm])
#     # if len(_pak8_services_to_apply) == 0:
#     #     pprint_status(f"No Pak8Services found")
#     #     return False
#     #
#     # pprint_status(f"Applying {len(_pak8_services_to_apply)} Pak8Services")
#     # pprint_status(f"Applying Namespace Resource")
#     # k8s_namespace_created = pak8_k8s_client.create_namespace_resource()
#     # pprint_status(f"Applying RBAC Resources")
#     # k8s_rbac_created = pak8_k8s_client.create_rbac_resources()
#     #
#     # k8s_namespace: str = pak8_k8s_client.get_k8s_namespace_to_use()
#     # k8s_sa: str = pak8_k8s_client.get_k8s_service_account_to_use()
#     # services_to_apply: List[pak8.Pak8ServiceBase] = []
#     # for _pak8_svc in _pak8_services_to_apply:
#     #     services_to_apply.append(
#     #        services.build_pak8_service_from_conf(
#     #             pak8_svc_conf=_pak8_svc,
#     #             namespace=k8s_namespace,
#     #             service_account=k8s_sa,
#     #             pak8_name=pak8_conf.name,
#     #             pak8_version=pak8_conf.version,
#     #             pak8_cloud_provider=pak8_conf.cloud_provider,
#     #         )
#     #     )
#     #
#     # k8s_api: Optional[pak8.K8sApi] = pak8_k8s_client.k8s_api
#     # for svc in services_to_apply:
#     #     svc.create_resources(
#     #         k8s_api=k8s_api,
#     #         namespace=k8s_namespace,
#     #         type_filters=type_filters,
#     #         name_filters=name_filters,
#     #     )
#     return True
#
#
# def get_active_k8s_resources_for_k8s_client(
#     pak8_conf: Pak8Conf,
#     pak8_k8s_client: Pak8K8sClient,
#     service_filters: Optional[List[str]] = None,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> Optional[Dict[str, List]]:
#     """For a given Pak8Conf and Pak8K8sClient, applies the k8s manifests
#
#     Args:
#     pak8_conf: Conf being deployed
#     pak8_k8s_client:
#
#     Returns:
#     bool: True is the resources were applied successfully
#     """
#
#     if is_empty(service_filters):
#         logger.debug("Getting Active K8s Resources")
#         return pak8_k8s_client.get_active_resources(
#             type_filters=type_filters,
#             name_filters=name_filters,
#         )
#
#     ######################################################
#     ## Get active resources for Pak8 Services
#     ######################################################
#
#     return None
#     # if pak8_conf.services is None:
#     #     return False
#     #
#     # # Create a list of services to apply
#     # _pak8_svc_dict = {s.name: s for s in pak8_conf.services}
#     # _pak8_services_to_scan: List[pak8.Pak8AppConf] = []
#     # for _svc_nm in service_filters:
#     #     if _svc_nm not in _pak8_svc_dict:
#     #         pprint_error(f"Service {_svc_nm} not found")
#     #         continue
#     #     _pak8_services_to_scan.append(_pak8_svc_dict[_svc_nm])
#     # if len(_pak8_services_to_scan) > 0:
#     #     pprint_status(
#     #         f"Getting Active K8s Resources for {len(_pak8_services_to_scan)} Pak8Services"
#     #     )
#     #
#     # k8s_namespace: str = pak8_k8s_client.get_k8s_namespace_to_use()
#     # k8s_sa: str = pak8_k8s_client.get_k8s_service_account_to_use()
#     # services_to_scan: List[pak8.Pak8ServiceBase] = []
#     # for _pak8_svc in _pak8_services_to_scan:
#     #     services_to_scan.append(
#     #        services.build_pak8_service_from_conf(
#     #             pak8_svc_conf=_pak8_svc,
#     #             namespace=k8s_namespace,
#     #             service_account=k8s_sa,
#     #             pak8_name=pak8_conf.name,
#     #             pak8_version=pak8_conf.version,
#     #             pak8_cloud_provider=pak8_conf.cloud_provider,
#     #         )
#     #     )
#     #
#     # k8s_api: Optional[pak8.K8sApi] = pak8_k8s_client.k8s_api
#     # active_k8s_resources: Dict[str, List] = defaultdict(list)
#     # for svc in services_to_scan:
#     #     _resources_for_svc = svc.get_active_resources(
#     #         k8s_api=k8s_api,
#     #         namespace=k8s_namespace,
#     #         type_filters=type_filters,
#     #         name_filters=name_filters,
#     #     )
#     #     for resource_type, active_objects in _resources_for_svc.items():
#     #         active_k8s_resources[resource_type].extend(active_objects)
#     # return active_k8s_resources



import datetime
from typing import Optional


def dttm_str_to_dttm(
    dttm_str: str, dttm_format: Optional[str] = None
) -> Optional[datetime.datetime]:
    """Convert a datestamp string to a Date object"""
    if dttm_str and dttm_str != "":
        dttm_formats = []
        if dttm_format:
            dttm_formats.append(dttm_format)
        else:
            dttm_formats.extend(
                [
                    "%Y-%m-%d",
                    "%Y-%m-%dT%H",
                    "%Y-%m-%d %H",
                    "%Y-%m-%dT%H:%M",
                    "%Y-%m-%d %H:%M",
                ]
            )
        for dttm_format in dttm_formats:
            try:
                datetime_object = datetime.datetime.strptime(dttm_str, dttm_format)
                return datetime_object
            except Exception:
                pass
    return None


def dttm_to_dttm_str(
    dttm: datetime.datetime, dttm_format: str = "%Y-%m-%dT%H:%M:%S"
) -> Optional[str]:
    """Convert a datetime object to formatted string"""

    if dttm and isinstance(dttm, datetime.datetime):
        try:
            dttm_str = datetime.datetime.strftime(dttm, dttm_format)
            return dttm_str
        except Exception as e:
            return None
    else:
        return None


def current_datetime_utc() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def yesterday_datetime() -> datetime.datetime:
    return datetime.datetime.now() - datetime.timedelta(days=1)



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Tuple

from phidata.asset.data_asset import DataAsset
from phidata.task.task import Task
from phidata.checks.check import Check
from phidata.product import DataProduct
from phidata.workflow import Workflow

from phiterm.workflow.exceptions import WorkflowNotFoundException
from phiterm.utils.cli_console import print_error, print_info, print_validation_errors
from phiterm.utils.log import logger


def parse_workflow_description(
    wf_description: str, ws_root_path: Path, workflows_dir: str
) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    This function parses a workflow description string in the format:
    Dir/File:Workflow:Task and return (workflow_file, workflow_name, task_name)

    The workflow_file is the path of the file starting after the workflows_dir.
    The workflow_file string can then be joined to the `workflows_dir` in the current env
    to get the absolute workflow file path
    Examples:

    `dau` -> (dau/dau.py, None)
    `dau:save` -> (dau/dau.py, save)
    `dau:save:test` -> (dau/dau.py, save, test)
    `dau/dau2:save:test` -> (dau/dau2.py, save, test)

    Args:
        wf_description:
        ws_root_path:
        workflows_dir:

    Returns:
        workflow_file: string path to the file containing the workflow,
                        relative to the products dir
        workflow_name: name of the workflow (if provided)
        task_name: name of the task (if provided)
    """

    # Split the wf_description by ":"
    wf_desc_slices = wf_description.strip().split(":")
    wf_path_slice: str = wf_desc_slices[0]
    wf_py_file: str = wf_path_slice + ".py"

    workflow_file: Optional[str] = None
    workflow_name: Optional[str] = (
        wf_desc_slices[1] if len(wf_desc_slices) > 1 else None
    )
    task_name: Optional[str] = wf_desc_slices[2] if len(wf_desc_slices) > 2 else None

    # Check if products dir is valid
    workflow_dir_path: Path = ws_root_path.joinpath(workflows_dir)
    if not (workflow_dir_path.exists() and workflow_dir_path.is_dir()):
        logger.error("Invalid workflow dir: {}".format(workflow_dir_path))
        return None, None, None

    # Check if the wf_path_slice is a file or a directory
    wf_py_file_path: Path = workflow_dir_path.joinpath(wf_py_file)
    wf_path_slice_path: Path = workflow_dir_path.joinpath(wf_path_slice)

    if wf_py_file_path.exists() and wf_py_file_path.is_file():
        # the path slice is a file
        workflow_file = wf_py_file
    elif wf_path_slice_path.exists() and wf_path_slice_path.is_dir():
        # the path slice is a dir and we should get the matching file.py
        wf_file_in_dir_path = wf_path_slice_path.joinpath(wf_py_file)
        if wf_file_in_dir_path.exists() and wf_file_in_dir_path.is_file():
            workflow_file = "{}/{}".format(wf_path_slice, wf_py_file)

    if workflow_file is None:
        logger.error("Workflow file not found")
        return None, None, None

    return workflow_file, workflow_name, task_name


def get_data_products_and_workflows_from_file(
    file_path: Path,
) -> Tuple[Dict[str, DataProduct], Dict[str, Workflow]]:
    """
    Reads the DataProducts & Workflows from filepath
    Args:
        file_path:

    Returns:

    """

    import importlib.util
    from importlib.machinery import ModuleSpec
    from pydantic import ValidationError

    logger.debug(f"Reading {file_path}")

    # Read DataProduct and Workflow objects from file_path
    try:
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        # Create a ModuleSpec
        dp_module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(
            "dp_module", file_path
        )
        # Using the ModuleSpec create a module
        if dp_module_spec:
            dp_module = importlib.util.module_from_spec(dp_module_spec)
            dp_module_spec.loader.exec_module(dp_module)  # type: ignore

            # loop over all objects in module and find DataProduct and Workflow objects
            dp_dict: Dict[str, DataProduct] = OrderedDict()
            wf_dict: Dict[str, Workflow] = OrderedDict()
            for k, v in dp_module.__dict__.items():
                if isinstance(v, DataProduct):
                    logger.debug(f"Found {k} | Type: {v.__class__.__name__}")
                    dp_dict[k] = v
                if isinstance(v, Workflow):
                    logger.debug(f"Found {k} | Type: {v.__class__.__name__}")
                    wf_dict[k] = v
            logger.debug("--^^-- Loading complete")
            return dp_dict, wf_dict
    except NameError as name_err:
        print_error("Variable not defined")
        raise
    except ValidationError as validation_err:
        print_error(str(validation_err))
        # print_validation_errors(validation_err.errors())
        print_info("Please fix and try again")
        exit(0)
    except (ModuleNotFoundError, Exception) as e:
        # logger.exception(e)
        raise
    raise WorkflowNotFoundException(f"No workflow found at {file_path}")


def parse_workflow_file(
    wf_file_path: Path,
) -> Tuple[
    Dict[str, DataProduct],
    Dict[str, Workflow],
    Dict[str, Task],
    Dict[str, DataAsset],
    Dict[str, Check],
]:
    """
    Reads a workflow_file and returns the
        - DataProducts
        - Workflows
        - Tasks
        - DataAssets
        - DQChecks
    Args:
        wf_file_path:

    Returns:

    """

    import importlib.util
    from importlib.machinery import ModuleSpec
    from pydantic import ValidationError

    logger.debug(f"Reading {wf_file_path}")

    # Read wf_file_path
    try:
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        # Create a ModuleSpec
        wf_module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(
            "wf_module", wf_file_path
        )
        # Using the ModuleSpec create a module
        if wf_module_spec:
            dp_module = importlib.util.module_from_spec(wf_module_spec)
            wf_module_spec.loader.exec_module(dp_module)  # type: ignore

            # loop over all objects in module and find DataProduct and Workflow objects
            dp_dict: Dict[str, DataProduct] = OrderedDict()
            wf_dict: Dict[str, Workflow] = OrderedDict()
            task_dict: Dict[str, Task] = OrderedDict()
            da_dict: Dict[str, DataAsset] = OrderedDict()
            dq_dict: Dict[str, Check] = OrderedDict()
            for k, v in dp_module.__dict__.items():
                if isinstance(v, DataProduct):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    dp_dict[v.name] = v
                if isinstance(v, Workflow):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    wf_dict[v.name] = v
                if isinstance(v, Task):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    task_dict[v.name] = v
                if isinstance(v, DataAsset):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    da_dict[v.name] = v
                if isinstance(v, Check):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    dq_dict[v.name] = v

            logger.debug("--^^-- Loading complete")
            return dp_dict, wf_dict, task_dict, da_dict, dq_dict
    except NameError as name_err:
        print_error("Variable not defined")
        raise
    except ValidationError as validation_err:
        print_error(str(validation_err))
        # print_validation_errors(validation_err.errors())
        print_info("Please fix and try again")
        exit(0)
    except ModuleNotFoundError as module_error:
        print_info("")
        print_error(str(module_error))
        print_info("Please install dependencies using `./scripts/install.sh`")
        exit(0)
    except FileNotFoundError as filenotfound_error:
        print_info("")
        print_error(str(filenotfound_error))
        exit(0)
    except Exception as e:
        # logger.exception(e)
        raise
    raise WorkflowNotFoundException(f"No workflow found at {wf_file_path}")



from pathlib import Path
from typing import Optional, Dict


def load_env(
    env: Optional[Dict[str, str]] = None, dotenv_dir: Optional[Path] = None
) -> None:
    from os import environ

    if dotenv_dir is not None:
        dotenv_file = dotenv_dir.joinpath(".env")
        if dotenv_file is not None and dotenv_file.exists() and dotenv_file.is_file():
            from dotenv.main import dotenv_values

            dotenv_dict: Dict[str, Optional[str]] = dotenv_values(dotenv_file)
            for key, value in dotenv_dict.items():
                if value is not None:
                    environ[key] = value

    if env is not None:
        environ.update(env)



import datetime
from pathlib import Path
from typing import List, Optional, Set

from pydantic import BaseModel
from phidata.workspace import WorkspaceConfig

from phiterm.schemas.workspace import WorkspaceSchema
from phiterm.utils.common import is_empty
from phiterm.utils.dttm import current_datetime_utc
from phiterm.utils.log import logger
from phiterm.workspace.ws_enums import WorkspaceSetupActions
from phiterm.workspace.exceptions import WorkspaceException


class PhiWsData(BaseModel):
    """The PhiWsData model stores data for a phidata workspace."""

    # Name of the workspace
    ws_name: str
    # WorkspaceSchema: If exists then indicates that this ws has been authenticated
    # with the backend
    ws_schema: Optional[WorkspaceSchema] = None
    # The root directory for the workspace.
    # This field indicates that this ws has been downloaded on this machine
    ws_root_path: Optional[Path] = None
    # DEPRECATED: Path for the WorkspaceConfig file
    ws_config_file_path: Optional[Path] = None
    # A Set of WorkspaceSetupActions which this workspace must satisfy to become valid.
    # if len(ws_data.required_actions.intersection(ws_data.pending_actions)) > 0:
    # the workspace is invalid
    required_actions: Set[WorkspaceSetupActions] = {
        WorkspaceSetupActions.WS_CONFIG_IS_AVL
    }
    # A Set of WorkspaceSetupActions which this workspace currently has not
    # completed and needs to fulfill.
    # When (pending_actions is None or len(pending_actions) == 0), the workspace is valid
    pending_actions: Set[WorkspaceSetupActions] = set()
    # Timestamp of when this workspace was created on the users machine
    create_ts: datetime.datetime = current_datetime_utc()
    # Timestamp when this ws was last updated
    last_update_ts: Optional[datetime.datetime] = None

    cached_ws_config: Optional[WorkspaceConfig] = None

    class Config:
        arbitrary_types_allowed = True

    ######################################################
    ## WorkspaceConfig functions
    ######################################################

    @property
    def ws_config(self) -> Optional[WorkspaceConfig]:
        """
        Returns the WorkspaceConfig for this workspace
        """

        if self.cached_ws_config is not None:
            return self.cached_ws_config

        if self.ws_root_path is None:
            raise WorkspaceException("Workspace root not set")

        from phiterm.workspace.ws_loader import add_ws_dir_to_path, load_workspace

        # NOTE: When loading a workspace, relative imports or package imports dont work.
        # This is a known problem in python :(
        #     eg: https://stackoverflow.com/questions/6323860/sibling-package-imports/50193944#50193944
        # To make them work, we need to install the workspace as a python module
        #   But when we run `phi ws setup`, the ws module is not yet installed
        # So we add workspace_root to sys.path so is treated as a module
        add_ws_dir_to_path(self.ws_root_path)

        logger.debug(f"**--> Loading WorkspaceConfig")
        loaded_config: WorkspaceConfig = load_workspace(ws_root_path=self.ws_root_path)

        # Update cached_ws_config
        self.cached_ws_config = loaded_config
        self.cached_ws_config.workspace_root_path = self.ws_root_path
        self.cached_ws_config.workspace_config_file_path = self.ws_config_file_path

        logger.debug("WorkspaceConfig:\n{}".format(self.cached_ws_config.args))
        return self.cached_ws_config

    ######################################################
    ## Print functions
    ######################################################

    def print_to_cli(self):
        from rich.pretty import pprint

        pprint(self.dict())
